@@ -15,19 +15,35 @@ def spherical_dist(lon_1,lat_1,lon_2,lat_2):
     return a*180/pi
 
 @jit(nopython=True)
-def in_rectangle(locs,alon,alat,blon,blat,width):
+def in_rectangle(locs,alon,alat,blon,blat,width,mode='normal'):
+    """
+    Judge a list of locations in within a certain distance of line a-b or not
+    mode: "normal" for Cartesian coordinate; "geo" for geographic coordinate
+    """
+    #----------quality control---------------------
+    assert len(locs.shape) == 2
+    assert mode in ["normal","geo"]
+
     results = np.zeros(locs.shape)
-    dlon1 = blon - alon
+    dlon1 = blon - alon                 # start point and end point
     dlat1 = blat - alat
-    rad_alon = radians(alon)
+    mlon1 = (blon+alon)/2
+    mlat1 = (blat+alat)/2
+    rad_alon = radians(alon)            # convert to radians
     rad_alat = radians(alat)
     norm1 = (dlon1**2+dlat1**2)**0.5
+#    if mode == "geo":
+#        norm1 = ((dlon1*np.cos(np.deg2rad((mlat1)))**2+dlat1**2)**0.5
     for i in range(locs.shape[0]):
         ilon = locs[i,0]
         ilat = locs[i,1]
         dlon2 = ilon - alon
         dlat2 = ilat - alat
+        mlon2 = (ilon+alon)/2
+        mlat2 = (ilat+alat)/2
         norm2 = (dlon2**2+dlat2**2)**0.5
+#    if mode == "geo":
+#        norm2 = ((dlon2*np.cos(np.deg2rad((mlat2)))**2+dlat2**2)**0.5
         if norm2 == 0:
             results[i,0]=1
             continue
@@ -89,7 +105,7 @@ def loc_by_width(lon1,lat1,lon2,lat2,width,direction='right'):
       lon2,lat2: longitude and latitude of tip 2
       direction: The side of new points from tip 1 to tip2 direction
     """
-    dlon = lon2 - lon1
+    dlon = (lon2 - lon1) * np.cos(np.deg2rad(lat1))
     dlat = lat2 - lat1
     dist=(dlon**2+dlat**2)**0.5
     
@@ -355,6 +371,9 @@ def sta_rotate(inFile,center,deg):
         f.write("\n")
 
 def spherical_rotate(lonlats,center,rotate):
+    """
+    rotate in degree, postive value for anti-clockwise direction
+    """
     lonlatBs = lonlats.copy()
     if isinstance(lonlatBs,list):
         lonlatBs = np.array(lonlatBs)
@@ -375,6 +394,8 @@ def _spherical_rotate(lonlatB,center,rotate):
     # C the rotated point
     # P the north pole
     # O the sphere center
+    
+    epsilon = 0.000001
 
     lonA = np.deg2rad(center[0])
     latA = np.deg2rad(center[1])
@@ -383,13 +404,13 @@ def _spherical_rotate(lonlatB,center,rotate):
     
     rotate = np.deg2rad(rotate)
     
-    if rotate % (2*pi) == 0:
-        lonC = lonB
-        latC = latB
-        return lonC,latC
+    if rotate % (2*pi) == 0:      # no change
+        return lonlatB[0],lonlatB[1]
 
     dist_deg = spherical_dist(center[0],center[1],lonlatB[0],lonlatB[1])
     AOB = np.deg2rad(dist_deg)
+    if AOB<epsilon:
+        return lonlatB[0],lonlatB[1]
 
     dlon = lonB - lonA
     dlat = latB - latA
@@ -399,7 +420,7 @@ def _spherical_rotate(lonlatB,center,rotate):
     
     if dlon>=0:
         tmp = sin(pi/2-latB)*sin(dlon)/sin(AOB)
-        if tmp >=1 and tmp <= 1.00000001:
+        if tmp >=1 and tmp <= 1.0+epsilon:
             tmp = 1
         if dlat >=0:
             PAB = asin(tmp)
@@ -414,13 +435,13 @@ def _spherical_rotate(lonlatB,center,rotate):
 
         # sines law
         tmp = sin(AOC)*sin(PAC)/sin(POC)
-        if tmp >=1 and tmp <= 1.00000001:
+        if tmp >=1 and tmp <= 1.0+epsilon:
             tmp = 1
         APC = asin(tmp)
         lonC = lonA + APC
     else:
         tmp = -sin(pi/2-latB)*sin(dlon)/sin(AOB)
-        if tmp >=1 and tmp <= 1.00000001:
+        if tmp >=1 and tmp <= 1.0+epsilon:
             tmp = 1
         if dlat >=0:
             PAB = asin(tmp)
@@ -435,7 +456,7 @@ def _spherical_rotate(lonlatB,center,rotate):
 
         # sines law
         tmp = sin(AOC)*sin(PAC)/sin(POC)
-        if tmp >=1 and tmp <= 1.00000001:
+        if tmp >=1 and tmp <= 1.0+epsilon:
             tmp = 1
         APC = asin(tmp)
         lonC = lonA - APC
@@ -443,16 +464,16 @@ def _spherical_rotate(lonlatB,center,rotate):
     BOC_deg = spherical_dist(np.rad2deg(lonB),np.rad2deg(latB),np.rad2deg(lonC),np.rad2deg(latC))
     BOC = np.deg2rad(BOC_deg)
     tmp = (cos(BOC)-cos(AOC)*cos(AOB))/(sin(AOC)*sin(AOB))
-    if tmp > 1 and tmp < 1.000001:
+    if tmp > 1 and tmp < 1.0+epsilon:
         tmp = 1
-    if tmp <-1 and tmp > -1.000001:
+    if tmp <-1 and tmp > -1.0-epsilon:
         tmp = -1
     inverse_rotate = acos(tmp)
     
     assert (np.abs(inverse_rotate) - np.abs(rotate)) <=0.01
     distAB = spherical_dist(np.rad2deg(lonA),np.rad2deg(latA),np.rad2deg(lonB),np.rad2deg(latB))
     distAC = spherical_dist(np.rad2deg(lonA),np.rad2deg(latA),np.rad2deg(lonC),np.rad2deg(latC))        
-    assert (distAB-distAC)<=0.000001
+    assert (distAB-distAC)<=epsilon
 
     return np.rad2deg(lonC),np.rad2deg(latC)
 
@@ -473,3 +494,74 @@ def mesh_rotate(x1,y1,center,rotate,method="Cartesian"):
     rotated_x1 = rotated_tmp_x1y1[:,0].reshape(ori_shape[0],ori_shape[1])
     rotated_y1 = rotated_tmp_x1y1[:,1].reshape(ori_shape[0],ori_shape[1])
     return rotated_x1,rotated_y1
+
+def fault_vectors(strike,dip,rake,unit='degree'):
+    """
+    rake: slip angle
+    return
+    | n: the unit normal vector of the fault plane
+    | d: the unit vector of the slip direction
+    | b: intermediate vector by n x d
+    | t: sigma3
+    | p: sigma1
+    """
+    if unit == 'degree':
+        strike = np.deg2rad(strike)
+        dip = np.deg2rad(dip)
+        rake = np.deg2rad(rake)
+    elif unit == 'radian':
+        pass
+    else:
+        raise Exception("Wrong unit, should be 'degree' or 'radian'")
+    
+    n = np.zeros(3)
+    n[0] = -1*np.sin(dip)*np.sin(strike)
+    n[1] = -1*np.sin(dip)*np.cos(strike)
+    n[2]=np.cos(dip)
+    
+    d = np.zeros(3)
+    d[0] = np.cos(rake)*np.cos(strike) + np.sin(rake)*np.cos(dip)*np.sin(strike)
+    d[1] = -1*np.cos(rake)*np.sin(strike)+np.sin(rake)*np.cos(dip)*np.cos(strike)
+    d[2] = np.sin(rake)*np.sin(dip)
+    
+    b = np.cross(n,d)
+    t = n+d
+    p = n-d
+    
+    return n,d,b,t,p
+
+def lonlat_by_dist(orglon,orglat,delta_x,delta_y,R=6378.1):
+    """
+    Assuming sphere earth, calculate new longitude and latitude 
+    base on delta x and y in kilometers.
+    
+    Parameters
+    |  orglon: Original longitude
+    |  Orglat: Oiginal latitude
+    | delta_x: distance along x direction in km
+    | delta_y: distance along y direction in km
+    """
+    # quality control
+    if orglat <-90 or orglat >90:
+        raise Exception("Latitude should be in [-90,90]")
+    if orglon<-180 or orglon>180:
+        raise Exception("Longitude should be in [-180,180]")
+    
+    delta_lat = delta_y/R/np.pi*180
+    newlat = orglat + delta_lat
+    if newlat < -90:
+        newlat =180 + newlat
+    if newlat >90:
+        newlat = 180-newlat
+        
+    R_ = R*np.cos(np.deg2rad(orglat))
+    delta_lon = delta_x/R_/np.pi*180
+    
+    newlon = orglon + delta_lon
+    
+    if newlon>180:
+        newlon -=180
+    if newlon <=-180:
+        newlon += 180
+
+    return newlon,newlat
