@@ -20,6 +20,7 @@ import logging
 import pickle
 from numba import cuda
 from distutils.sysconfig import get_python_lib
+import matplotlib.pyplot as plt
 
 def add_path():
     print(">>> Add path to python library ...")
@@ -425,3 +426,186 @@ def writefile(cont,texfile):
     with open(texfile,'w') as f:
         for line in cont:
             f.write(line+"\n")
+
+def bdy2pts(xmin,xmax,ymin,ymax):
+    """
+    Convert boundary to points for plot
+    """
+    points = []
+    points.append([xmin,ymin])
+    points.append([xmin,ymax])
+    points.append([xmax,ymax])
+    points.append([xmax,ymin])
+    points.append([xmin,ymin])
+    return np.array(points)
+def matrix_show(*args,**kwargs):
+    """
+    Show matrix values in grids shape
+    Parameters:cmap="cool",gridsize=0.6,fmt='.2f',label_data=True
+    """
+    ws = []
+    H = 0
+    str_count = 0
+    ndarr_count = 0
+    new_args = []
+    for arg in args:
+        if isinstance(arg,str):
+            new_args.append(arg)
+            continue
+        if isinstance(arg,list):
+            arg = np.array(arg)
+        if len(arg.shape)>2:
+            raise Exception("Only accept 2D array")
+        if len(arg.shape) == 1:
+            n = arg.shape[0]
+            tmp = np.zeros((n,1))
+            tmp[:,0] = arg.ravel()
+            arg = tmp
+        h,w = arg.shape
+        if h>H:
+            H=h
+        ws.append(w)
+        new_args.append(arg)
+        ndarr_count += 1
+    W = np.sum(ws)+len(ws)    # text+matrix+text+...+matrix+text
+    if W<0:
+        raise Exception("No matrix provided!")
+        
+    fmt = '.2f'
+    grid_size = 0.6
+    cmap = 'cool'
+    label_data = True
+    for arg in kwargs:
+        if arg == "fmt":
+            fmt = kwargs[arg]
+        if arg == 'grid_size':
+            grid_size = kwargs[arg]
+        if arg == 'cmap':
+            cmap = kwargs[arg]
+        if arg == 'label_data':
+            label_data = kwargs[arg]
+    fig = plt.figure(figsize=(W*grid_size,H*grid_size))
+    gs = fig.add_gridspec(nrows=H,ncols=W)
+    
+    wloop = 0
+    matrix_id = 0
+    for arg in new_args:
+        if isinstance(arg,str):
+            ax = fig.add_subplot(gs[0:H,wloop-1:wloop])
+            ax.axis("off")
+            ax.set_xlim(0,1)
+            ax.set_ylim(0,H)
+            ax.text(0.5,H/2,arg,horizontalalignment='center',verticalalignment='center')
+        if isinstance(arg,np.ndarray):
+            h,w = arg.shape
+            hlow = int(np.round((H-h+0.01)/2))        # Find the height grid range
+            hhigh = hlow+h
+            wlow = wloop
+            whigh = wlow+w
+#            print("H: ",H,hlow,hhigh,"; W ",W,wlow,whigh)
+            ax = fig.add_subplot(gs[hlow:hhigh,wlow:whigh])
+            
+            plt.pcolormesh(arg,cmap=cmap)
+            for i in range(1,w):
+                plt.axvline(i,color='k',linewidth=0.5)
+            for j in range(1,h):
+                plt.axhline(j,color='k',linewidth=0.5)
+            if label_data:
+                for i in range(h):
+                    for j in range(w):
+                        plt.text(j+0.5,i+0.5,format(arg[i,j],fmt),
+                                 horizontalalignment='center',
+                                 verticalalignment='center')
+            plt.xlim(0,w)
+            plt.ylim([h,0])
+            plt.xticks([])
+            plt.yticks([])
+            wloop+=w+1
+            matrix_id+=1
+    plt.show()
+
+def layer2linear_vel(inp_depths,inp_vels,linear_nodes):
+    """
+    Convert from layered 1-D model to linear velocity model, designed to conver from 
+    the HYPOINVERSE model to tomoDD input model. It balances the average slowness.
+    Parameters:
+    | inp_depths: input depths, 1-D array/list
+    | inp_vels: input velocities, 1-D array/list
+    | linear_nodes: the depth nodes of output velocities
+    
+    Return:
+    | average vlocity list correponding to the input velocity nodes
+
+    Example:
+    >>> inp_depths = [ -1,0.5,1.,1.5,2.13,4.13,6.9,10.07,15.93,17.,18.,27.,31.89]
+    >>> inp_vels = [3.67,4.7,5.38,5.47,5.5,5.61,5.73,6.19,6.23,6.31,6.4,6.45,6.5]
+    >>> tomo_deps = [0,1.5,3.0,4.5,6.0,7.5,9.0,10.5,12,18,24]
+    >>> layer2linear_vel(inp_depths,inp_vels,tomo_deps)
+    """
+    
+    #----------- quality control --------------------------------
+    if linear_nodes[-1]>inp_depths[-1]:
+        raise Exception("linear_nodes[-1] should less than inp_depths[-1]")
+    if inp_depths.shape[0] != inp_vels.shape[0]:
+        raise Eception("Different length of input depths quantity and velocity values")
+    
+    #----------- load in data ------------------------------------
+    pxnodes = []
+    pynodes = []
+    for i in range(inp_depths.shape[0]):
+        if i == 0:
+            pxnodes.append(inp_vels[i])
+            pynodes.append(inp_depths[i])
+        else:
+            pxnodes.append(inp_vels[i-1])
+            pxnodes.append(inp_vels[i])
+            pynodes.append(inp_depths[i])
+            pynodes.append(inp_depths[i])
+
+    #----------- calculate the mean velocity ----------------------
+    avg_vels = []
+    for i in range(len(linear_nodes)):
+        dep = linear_nodes[i]
+        #---------------get left and right range--------------
+        if i == 0:                                   # first node
+            dep_il = linear_nodes[i]
+            dep_ir = (linear_nodes[i+1]+linear_nodes[i])/2
+        elif i == len(linear_nodes) - 1:             # last node
+            dep_il = (linear_nodes[i-1]+linear_nodes[i])/2
+            dep_ir = linear_nodes[i]+(linear_nodes[i]-linear_nodes[i-1])/2
+        else:
+            dep_il = (linear_nodes[i-1]+linear_nodes[i])/2
+            dep_ir = (linear_nodes[i+1]+linear_nodes[i])/2
+
+        slw_list = []                                # slow list
+        embrace_first = False
+        for j in range(len(pynodes)):
+            pynode = pynodes[j]
+            velocity = pxnodes[j]                    # velocity
+
+            if pynode >=dep_il: # consider velocity node within the range
+                if embrace_first == False:
+                    slw_list.append([dep_il,1/velocity]) # embrace the left node
+                    embrace_first = True
+
+                end_velocity = velocity
+                if pynode>dep_ir:
+                    break
+                slw_list.append([pynode,1/velocity])
+        slw_list.append([dep_ir,1/end_velocity])      # embrace the right node
+
+        # calculate average slowness
+        trav = 0                                      # travel time        
+        for k in range(len(slw_list)-1):
+            dep_a = slw_list[k][0]
+            slw_a = slw_list[k][1]
+            dep_b = slw_list[k+1][0]
+            slw_b = slw_list[k+1][1]
+            if (dep_b - dep_a)==0 or slw_a==slw_b:
+                trav += (dep_b -dep_a)*slw_a
+            else:
+                raise Exception("(dep_b - dep_a)==0 or vel_a==vel_b")
+        avg_vel = (dep_ir-dep_il)/trav
+        avg_vels.append(avg_vel)
+        
+    return avg_vels
